/*
Based on example.c from libpng by Guy Eric Schalnat.
Any errors or mistakes are Smarasderagd's fault.
*/

#include "xli.h"
#include "pbm.h"
#include <png.h>
#include "assert.h"

#define TITLE_KEYWORD "Title"

/* check to see if a file is a png file using png_check_sig() */
static int check_png(file_name)
char *file_name;
{
    ZFILE *zfp;
    char buf[8];
    int ret;

    zfp = zopen(file_name, "rb");
    if (!zfp)
        return 0;
    ret = zread(zfp, buf, 8);
    zclose(zfp);

    if (ret != 8)
        return 0;

    ret = png_check_sig(buf, 8);

    return (ret);
}


static int describe_png(name, pngp, info)
char *name;
png_struct *pngp;
png_info *info;
{
    printf("%s is a%s %ux%u %d bit deep %s PNG image",
        name,
        info->interlace_type ? "n interlaced" : "",
        info->width, info->height, info->bit_depth,
        (info->color_type & PNG_COLOR_MASK_PALETTE) ? "paletted" :
            (info->color_type & PNG_COLOR_MASK_COLOR) ? "RGB" : "grayscale");
    if (info->color_type & PNG_COLOR_MASK_ALPHA)
        printf(" with an alpha channel");
    if (info->valid & PNG_INFO_tRNS)
        printf(" with transparency");
    if (info->valid & PNG_INFO_bKGD) {
        png_color_16 *bg = &(info->background);
        int d = (info->bit_depth + 3) / 4;

        printf(", background = ");
        if (!(PNG_COLOR_MASK_COLOR & info->color_type)) {
            printf("#%0*x", d, bg->gray);
        } else if (PNG_COLOR_TYPE_PALETTE == info->color_type) {
            png_color *bp = info->palette + bg->index;
            printf("index %d: #%02x%02x%02x", bg->index,
                bp->red, bp->green, bp->blue);
        } else {
            printf("#%0*x%0*x%0*x", d, bg->red, d, bg->green, d, bg->blue);
        }
    }
    if (info->valid & PNG_INFO_gAMA)
        printf(", gamma = %g", info->gamma);
    printf("\n");
}


static void xli_png_read_data(pngp, data, length)
png_struct *pngp;
png_byte *data;
png_uint_32 length;
{
    ZFILE *zfp;

    zfp = (ZFILE *) png_get_io_ptr(pngp);
    if (zread(zfp, (byte *) data, (int) length) != length)
        png_error(pngp, "unexpected EOF in xli_png_read_data");
}


Image *pngLoad(fullname, opt, verbose)
ImageOptions *opt;
char *fullname;
boolean verbose;
{
    ZFILE *zfp;
    png_struct *pngp;
    png_info *info;
    Image *image;
    int orig_depth = 0;
    byte **row = (byte **) 0;

    /* open the file */
    if (!check_png(fullname))
        return (Image *) 0;

    zfp = zopen(fullname);
    if (!zfp) {
        perror("pngLoad");
        return (Image *) 0;
    }

    /* allocate the necessary structures */
    pngp = (png_struct *) lmalloc(sizeof (png_struct));
    if (!pngp)
    {
        zclose(zfp);
        return (Image *) 0;
    }

    info = (png_info *) lmalloc(sizeof (png_info));
    if (!info)
    {
        zclose(zfp);
        lfree(pngp);
        return (Image *) 0;
    }

    /* set error handling */
    if (setjmp(pngp->jmpbuf))
    {
        png_read_destroy(pngp, info, (png_info *)0);
        zclose(zfp);
        lfree(pngp);
        lfree(info);
        if (row)
            lfree(row);
        /* If we get here, we had a problem reading the file */
        return (Image *) 0;
    }

    /* initialize the structures, info first for error handling */
    png_info_init(info);
    png_read_init(pngp);

    /* set up the input control */
    png_set_read_fn(pngp, (void *) zfp, xli_png_read_data);

    /* read the file information */
    png_read_info(pngp, info);

    if (verbose)
        describe_png(fullname, pngp, info);

    znocache(zfp);

    /* the rules:
        bit depth over 8        ITRUE
        rgb colour              ITRUE
        1 bit gray              IBITMAP
        other gray              IRGB
        palette colour          IRGB
    */

    if (info->bit_depth < 8 && (PNG_COLOR_TYPE_RGB == info->color_type ||
            PNG_COLOR_TYPE_RGB_ALPHA == info->color_type))
        png_set_expand(pngp);

    /* Set the background color to draw transparent and alpha images over */
    if ((info->color_type & PNG_COLOR_MASK_ALPHA) ||
            ((info->valid & PNG_INFO_tRNS) &&
            (opt->bg || (info->valid & PNG_INFO_bKGD)))) {
        png_color_16 bg;
        int expand = 0;
        int gflag = PNG_BACKGROUND_GAMMA_SCREEN;
        double gval = 1.0;

        bg.red = bg.green = bg.blue = bg.gray = 0;
        if (PNG_COLOR_TYPE_PALETTE == info->color_type || opt->bg)
            png_set_expand(pngp);

        if (opt->bg) {
            xliXColor xc;
            int shift = ((info->color_type & PNG_COLOR_MASK_ALPHA) &&
                    16 == info->bit_depth) ? 0 : 8;

            xc.flags= DoRed | DoGreen | DoBlue;
            xliParseXColor(&(globals.dinfo), opt->bg, &xc);
            bg.red = xc.red >> shift;
            bg.green = xc.green >> shift;
            bg.blue = xc.blue >> shift;
            bg.gray = (int) (xc.red * .299 + xc.green * .587 + xc.blue * .114
                + .5) >> shift;
        } else if (info->valid & PNG_INFO_bKGD) {
            bg = info->background;
            expand = 1;
            gflag = PNG_BACKGROUND_GAMMA_FILE;
        }
        png_set_background(pngp, &bg, gflag, expand, gval);
    }

    if (info->bit_depth < 8 && (info->bit_depth > 1 ||
            PNG_COLOR_TYPE_GRAY != info->color_type)) {
        if (PNG_COLOR_TYPE_GRAY == info->color_type)
            orig_depth = info->bit_depth;
        png_set_packing(pngp);
    }
 
    /* tell libpng to strip 16 bit depth files down to 8 bits */
    if (info->bit_depth > 8)
        png_set_strip_16(pngp);

    /* AUGH */
    png_set_interlace_handling(pngp);

    /* update palette with transformations, update the info structure */
    png_read_update_info(pngp, info);

    /* allocate the memory to hold the image using the fields
        of png_info. */
    if (PNG_COLOR_TYPE_GRAY == info->color_type && 1 == info->bit_depth) {
        image = newBitImage(info->width, info->height);
        png_set_invert_mono(pngp);
    } else if (PNG_COLOR_TYPE_PALETTE == info->color_type) {
        int i;

        image = newRGBImage(info->width, info->height, info->bit_depth);
        for (i = 0; i < info->num_palette; ++i) {
            image->rgb.red[i] = info->palette[i].red << 8;
            image->rgb.green[i] = info->palette[i].green << 8;
            image->rgb.blue[i] = info->palette[i].blue << 8;
        }
        image->rgb.used = info->num_palette;
    } else if (PNG_COLOR_TYPE_GRAY == info->color_type) {
        int i;
        int depth = orig_depth ? orig_depth : info->bit_depth;
        int maxval = (1 << depth) - 1;

        image = newRGBImage(info->width, info->height, depth);
        for (i = 0; i <= maxval; i++) {
            image->rgb.red[i] = PM_SCALE(i, maxval, 0xffff);
            image->rgb.green[i] = PM_SCALE(i, maxval, 0xffff);
            image->rgb.blue[i] = PM_SCALE(i, maxval, 0xffff);
        }
        image->rgb.used = maxval + 1;
    } else {
        image = newTrueImage(info->width, info->height);
    }

    if (info->valid & PNG_INFO_gAMA && image->type != IBITMAP)
        image->gamma = 1.0 / info->gamma;

    /* read the image */
    if (IBITMAP == image->type)
        assert((image->width + 7) / 8 == info->rowbytes);
    else
        assert(image->width * image->pixlen == info->rowbytes);

    {
        int i;

        row = (byte **) lmalloc(info->height * sizeof(byte *));
        for (i = 0; i < info->height; ++i)
            row[i] = image->data + i * info->rowbytes;
    }

    png_read_image(pngp, row);

    /* read the rest of the file, getting any additional chunks in info */
    png_read_end(pngp, info);

    if (verbose && (info->valid & PNG_INFO_tIME)) {
        printf("Modification time: %d/%d/%d %02d:%02d:%02d",
            info->mod_time.year, info->mod_time.month, info->mod_time.day,
            info->mod_time.hour, info->mod_time.minute, info->mod_time.second);
    }

    {
        int i;
        char *title;

        title = opt->name;
        for (i = 0; i < info->num_text; ++i) {
            if (verbose)
                printf("\n%s: %s\n", info->text[i].key, info->text[i].text);
            if (!strcmp(info->text[i].key, TITLE_KEYWORD))
                title = info->text[i].text;
        }
        image->title = dupString(title);
    }

    /* clean up after the read, and free any memory allocated */
    png_read_destroy(pngp, info, (png_info *)0);

    /* free the structures */
    lfree(pngp);
    lfree(info);
    lfree(row);

    /* close the file */
    zclose(zfp);

    /* that's it */
    return image;
}


int pngIdent(fullname,name)
char *fullname;
char *name;
{
    ZFILE *zfp;
    png_struct *pngp;
    png_info *info;
    int i;

    if (!check_png(fullname))
        return 0;

    /* open the file */
    zfp = zopen(fullname);
    if (!zfp) {
        perror("pngIdent");
        return 0;
    }

    /* allocate the necessary structures */
    pngp = (png_struct *) lmalloc(sizeof (png_struct));
    if (!pngp)
    {
        zclose(zfp);
        return 0;
    }

    info = (png_info *) lmalloc(sizeof (png_info));
    if (!info)
    {
        zclose(zfp);
        lfree(pngp);
        return 0;
    }

    /* set error handling */
    if (setjmp(pngp->jmpbuf))
    {
        png_read_destroy(pngp, info, (png_info *)0);
        zclose(zfp);
        lfree(pngp);
        lfree(info);
        /* If we get here, we had a problem reading the file */
        return 0;
    }

    /* initialize the structures, info first for error handling */
    png_info_init(info);
    png_read_init(pngp);

    /* set up the input control */
    png_set_read_fn(pngp, (void *) zfp, xli_png_read_data);

    /* read the file information */
    png_read_info(pngp, info);

    describe_png(fullname, pngp, info);

    /* clean up after the read, and free any memory allocated */
    png_read_destroy(pngp, info, (png_info *) 0);

    /* free the structures */
    lfree(pngp);
    lfree(info);

    /* close the file */
    zclose(zfp);

    /* that's it */
    return 1;
}
